#include <stdlib.h>
#include <string.h>
#include <libintl.h>
#include <gtk/gtk.h>
#include <gdk/gdk.h>
#include <gdk/gdkkeysyms.h>
#include <m17n.h>

static MSymbol current_lang, current_name, current_command;

static MPlist *
get_command_bindings (MSymbol lang, MSymbol name, MSymbol command)
{
  return (mplist_next
	  (mplist_next
	   (mplist_next
	    (mplist_value
	     (minput_get_command (lang, name, command))))));
}

static int
keyseq_equal (MPlist *pl1, MPlist *pl2)
{
  if (mplist_length (pl1) != mplist_length (pl2))
    return 0;
  while (pl1 && mplist_key (pl1) == Msymbol)
    {
      if (mplist_value (pl1) != mplist_value (pl2))
	return 0;
      pl1 = mplist_next (pl1);
      pl2 = mplist_next (pl2);
    }
  return 1;
}

static void
keyseq_render_function (GtkTreeViewColumn *column,
			GtkCellRenderer *renderer,
			GtkTreeModel *model,
			GtkTreeIter *iter,
			gpointer data)
{
  MPlist *keyseq, *pl;
  gint n;
  gchar buf[1024];

  gtk_tree_model_get (model, iter, 0, &keyseq, -1);
  for (pl = keyseq, n = 0;
       pl && mplist_key (pl) == Msymbol;
       pl = mplist_next (pl))
    n += strlen (msymbol_name ((MSymbol) mplist_value (pl))) + 1;
  if (n < sizeof (buf))
    {
      buf[0] = '\0';
      for (pl = keyseq;
	   pl && mplist_key (pl) == Msymbol;
	   pl = mplist_next (pl))
	{
	  strcat (buf, msymbol_name ((MSymbol) mplist_value (pl)));
	  strcat (buf, " ");
	}
      g_object_set (renderer, "foreground-set", FALSE, NULL);
    }
  else
    {
      g_snprintf (buf, sizeof (buf), "Too long to display");
      g_object_set (renderer, "foreground", "Red", "foreground-set", TRUE,
		    NULL);
    }
  g_object_set (renderer, "text", buf, NULL);
}

static void
delete_cb (GtkButton *button, GtkWidget *view)
{
  GtkTreeSelection *selection;
  GtkTreeModel *model;
  GtkTreeIter iter;
  MPlist *keyseq, *old, *new, *pl;

  selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (view));
  if (gtk_tree_selection_get_selected (selection, &model, &iter))
    {
      gtk_tree_model_get (model, &iter, 0, &keyseq, -1);

      old = get_command_bindings (current_lang, current_name, current_command);
      new = mplist ();
      for (pl = old; pl && mplist_key (pl) == Mplist; pl = mplist_next (pl))
	if (! keyseq_equal (mplist_value (pl), keyseq))
	  mplist_add (new, Mplist, mplist_value (pl));
      minput_config_command (current_lang, current_name, current_command, new);
      m17n_object_unref (new);
      gtk_list_store_remove (GTK_LIST_STORE (model), &iter);
    }
}
  

static void
delete_keyseq (GtkWidget *view, MSymbol lang, MSymbol name, MSymbol command)
{
  GtkTreeSelection *selection;
  GtkTreeModel *model;
  GtkTreeIter iter;
  MPlist *keyseq, *old, *new, *pl;

  selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (view));
  if (gtk_tree_selection_get_selected (selection, &model, &iter))
    {
      gtk_tree_model_get (model, &iter, 0, &keyseq, -1);
      old = get_command_bindings (lang, name, command);
      new = mplist ();
      for (pl = old;
	   pl && mplist_key (pl) == Mplist;
	   pl = mplist_next (pl))
	if (! keyseq_equal (mplist_value (pl), keyseq))
	  mplist_add (new, Mplist, mplist_value (pl));
      minput_config_command (lang, name, command, new);
      m17n_object_unref (new);
      gtk_list_store_remove (GTK_LIST_STORE (model), &iter);
    }
}

static MPlist *entry_keyseq;

static void
update_entry (GtkEntry *entry)
{
  if (mplist_key (entry_keyseq) == Mnil)
    gtk_entry_set_text (entry, "");
  else
    {
      MPlist *p;
      gchar *name;

      name = msymbol_name ((MSymbol) mplist_value (entry_keyseq));
      gtk_entry_set_text (entry, name);
      for (p = mplist_next (entry_keyseq); mplist_key (p) != Mnil;
	   p = mplist_next (p))
	{
	  name = msymbol_name ((MSymbol) mplist_value (p));
	  gtk_entry_append_text (entry, " ");
	  gtk_entry_append_text (entry, name);
	}
      gtk_editable_set_position (GTK_EDITABLE (entry), -1);
    }
}

enum KeyMaskBit {
  META_MASK_BIT = 1,
  ALT_MASK_BIT = META_MASK_BIT << 1,
  SUPER_MASK_BIT = ALT_MASK_BIT << 1,
  HYPER_MASK_BIT = SUPER_MASK_BIT << 1
};

static unsigned modifier_state = 0;

gboolean
key_pressed (GtkEntry *entry, GdkEventKey *event, gpointer data)
{
  guint c;
  MText *mt;
  char buf[32];
  char *name;
  int nbytes, i;

  c = gdk_keyval_to_unicode (event->keyval);
  if (c == 0)
    {
      switch (event->keyval)
	{
	case GDK_Meta_L: case GDK_Meta_R:
	  modifier_state |= META_MASK_BIT; return TRUE;
	case GDK_Alt_L: case GDK_Alt_R:
	  modifier_state |= ALT_MASK_BIT; return TRUE;
	case GDK_Super_L: case GDK_Super_R:
	  modifier_state |= SUPER_MASK_BIT; return TRUE;
	case GDK_Hyper_L: case GDK_Hyper_R:
	  modifier_state |= HYPER_MASK_BIT; return TRUE;
	default:
	  if (event->keyval >= GDK_Shift_L && event->keyval <= GDK_Shift_Lock)
	    return TRUE;
	}
      name = gdk_keyval_name (event->keyval);
      if (! name)
	return TRUE;
      nbytes = strlen (name);
    }
  else
    {
      name = alloca (8);
      mt = mtext ();
      mtext_cat_char (mt, c);
      nbytes = mconv_encode_buffer (msymbol ("utf-8"), mt,
				    (unsigned char *) name, 32);
      m17n_object_unref (mt);
    }
  i = 0;
  if (c == 0 && event->state & GDK_SHIFT_MASK)
    buf[i++] = 'S', buf[i++] = '-';
  if (event->state & GDK_CONTROL_MASK)
    buf[i++] = 'C', buf[i++] = '-';
  if (modifier_state & META_MASK_BIT)
    buf[i++] = 'M', buf[i++] = '-';
  if (modifier_state & ALT_MASK_BIT)
    buf[i++] = 'A', buf[i++] = '-';
  if (modifier_state & SUPER_MASK_BIT)
    buf[i++] = 's', buf[i++] = '-';
  if (modifier_state & HYPER_MASK_BIT)
    buf[i++] = 'H', buf[i++] = '-';
  strncpy (buf + i, name, nbytes);
  buf[i + nbytes] = 0;
  mplist_add (entry_keyseq, Msymbol, msymbol (buf));
  update_entry (entry);
  return TRUE;
}

gboolean
key_released (GtkEntry *entry, GdkEventKey *event, gpointer data)
{
  guint c;

  c = gdk_keyval_to_unicode (event->keyval);
  if (c == 0)
    {
      switch (event->keyval)
	{
	case GDK_Meta_L: case GDK_Meta_R:
	  modifier_state &= ~META_MASK_BIT; break;
	case GDK_Alt_L: case GDK_Alt_R:
	  modifier_state &= ~ALT_MASK_BIT; break;
	case GDK_Super_L: case GDK_Super_R:
	  modifier_state &= ~SUPER_MASK_BIT; break;
	case GDK_Hyper_L: case GDK_Hyper_R:
	  modifier_state &= ~HYPER_MASK_BIT; break;
	}
    }
  return FALSE;
}

static void
clear_cb (GtkButton *button, GtkWidget *entry)
{
  mplist_set (entry_keyseq, Mnil, NULL);
  gtk_widget_grab_focus (entry);
  update_entry (GTK_ENTRY (entry));
}

static void
add_cb (GtkButton *button, GtkListStore *store)
{
  MPlist *old, *new, *pl, *last;
  GtkTreeIter iter;
  GtkWidget *view;

  if (mplist_length (entry_keyseq) == 0)
    return;
  old = get_command_bindings (current_lang, current_name, current_command);
  new = mplist ();
  for (pl = old; pl && mplist_key (pl) == Mplist; pl = mplist_next (pl))
    {
      if (! keyseq_equal (mplist_value (pl), entry_keyseq))
	mplist_add (new, Mplist, mplist_value (pl));
      else
	{
	  /* entry_keyseq is already registered. */
	  m17n_object_unref (new);
	  return;
	}
    }
  mplist_add (new, Mplist, entry_keyseq);
  minput_config_command (current_lang, current_name, current_command, new);
  m17n_object_unref (new);

  /* We cannot use ENTRY_KEYSEQ for gtk_list_store_set ().  We must
     use a pointer to the internally copied one. */
  new = get_command_bindings (current_lang, current_name, current_command);
  for (pl = new; pl && mplist_key (pl) == Mplist;
       last = pl, pl = mplist_next (pl))
  gtk_list_store_append (store, &iter);
  gtk_list_store_set (store, &iter, 0, mplist_value (last), -1);
  view = gtk_tree_view_new_with_model (GTK_TREE_MODEL (store));
}

static void
command_cb (GtkButton *button, GtkWidget *frame)
{
  GtkWidget *dialog, *view, *scrolled;
  GtkWidget *delete, *entry, *add, *clear;
  GtkListStore *store;
  GtkTreeIter iter;
  GtkTreeViewColumn *column;
  GtkCellRenderer *renderer;
  MPlist *bindings, *pl, *cmd;
  gint response;

  current_lang = g_object_get_data (G_OBJECT (button), "mim_lang");
  current_name = g_object_get_data (G_OBJECT (button), "mim_name");
  current_command = g_object_get_data (G_OBJECT (button), "mim_command");
  bindings = get_command_bindings (current_lang, current_name, current_command);

  dialog = (gtk_dialog_new_with_buttons
	    (msymbol_name (current_command),
	     GTK_WINDOW (gtk_widget_get_toplevel (frame)),
	     GTK_DIALOG_DESTROY_WITH_PARENT | GTK_DIALOG_NO_SEPARATOR,
	     GTK_STOCK_OK, GTK_RESPONSE_OK,
	     GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
	     NULL));

  store = gtk_list_store_new (1, G_TYPE_POINTER);
  for (pl = bindings; pl && mplist_key (pl) == Mplist; pl = mplist_next (pl))
    {
      gtk_list_store_append (store, &iter);
      gtk_list_store_set (store, &iter, 0, mplist_value (pl), -1);
    }
  view = gtk_tree_view_new_with_model (GTK_TREE_MODEL (store));
  g_object_unref (G_OBJECT (store));
  column = gtk_tree_view_column_new ();
  gtk_tree_view_column_set_title (column, "Current bindings");
  gtk_tree_view_append_column (GTK_TREE_VIEW (view), column);
  renderer = gtk_cell_renderer_text_new ();
  gtk_tree_view_column_pack_start (column, renderer, TRUE);
  gtk_tree_view_column_set_cell_data_func
    (column, renderer, keyseq_render_function, NULL, NULL);
  gtk_container_add (GTK_CONTAINER (GTK_DIALOG (dialog)->vbox), view);

  delete = gtk_button_new_from_stock (GTK_STOCK_DELETE);
  g_signal_connect (G_OBJECT (delete), "clicked",
		    G_CALLBACK (delete_cb), view);
  gtk_container_add (GTK_CONTAINER (GTK_DIALOG (dialog)->vbox), delete);

  entry_keyseq = mplist ();

  entry = gtk_entry_new ();
  g_signal_connect (G_OBJECT (entry), "key-press-event",
		    G_CALLBACK (key_pressed), NULL);
  g_signal_connect (G_OBJECT (entry), "key-release-event",
		    G_CALLBACK (key_released), NULL);
  gtk_container_add (GTK_CONTAINER (GTK_DIALOG (dialog)->vbox), entry);

  clear = gtk_button_new_from_stock (GTK_STOCK_CLEAR);
  g_signal_connect (G_OBJECT (clear), "clicked",
		    G_CALLBACK (clear_cb), entry);
  gtk_container_add (GTK_CONTAINER (GTK_DIALOG (dialog)->vbox), clear);

  add = gtk_button_new_from_stock (GTK_STOCK_ADD);
  g_signal_connect (G_OBJECT (add), "clicked",
		    G_CALLBACK (add_cb), store);
  gtk_container_add (GTK_CONTAINER (GTK_DIALOG (dialog)->vbox), add);

  gtk_widget_show_all (dialog);
  response = gtk_dialog_run (GTK_DIALOG (dialog));
  gtk_widget_destroy (dialog);

  m17n_object_unref (entry_keyseq);
}

void
create_command_entries (GtkWidget *frame, GtkTooltips *tip,
			MSymbol lang, MSymbol name)
{
  GtkWidget *vbox;
  MPlist *plist;

  plist = minput_get_command (lang, name, Mnil);
  if (! plist)
    {
      GtkWidget *message;

      message = gtk_label_new ("No commands for this method");
      gtk_container_add (GTK_CONTAINER (frame), message);
      return;
    }

  /*
   * plist == ((command description status keyseq keyseq ...)
   *           (command description status keyseq keyseq ...)
   *           ...)
   */
  vbox = gtk_vbox_new (FALSE, 0);
  gtk_container_add (GTK_CONTAINER (frame), vbox);

  for (; plist && mplist_key (plist) == Mplist; plist = mplist_next (plist))
    {
      GtkWidget *button;
      MPlist *pl;
      MSymbol command;
      gchar *desc;

      pl = mplist_value (plist);
      /* pl == (command description status keyseq keyseq ...) */
      command = mplist_value (pl);
      button = gtk_button_new_with_label (msymbol_name (command));
      gtk_container_add (GTK_CONTAINER (vbox), button);

      pl = mplist_next (pl);
      /* pl == (description status keyseq keyseq ...) */
      if (mplist_key (pl) == Mtext)
	/* Fixme : Assuming the return value is in UTF-8 */
	desc = mtext_data (mplist_value (pl), NULL, NULL, NULL, NULL);
      else
	desc = NULL;
      gtk_tooltips_set_tip (tip, button, desc, NULL);

      g_object_set_data (G_OBJECT (button), "mim_lang", lang);
      g_object_set_data (G_OBJECT (button), "mim_name", name);
      g_object_set_data (G_OBJECT (button), "mim_command", command);
      g_signal_connect (G_OBJECT (button), "clicked",
			G_CALLBACK (command_cb), frame);
    }
  m17n_object_unref (plist);
}

