/* mim-config.c -- M17N input method configuration
   Copyright (C) 2007
     National Institute of Advanced Industrial Science and Technology (AIST)
     Registration Number H15PRO112

   This file is part of the m17n-im-config package; a sub-part of the
   m17n library.

   The m17n library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public License
   as published by the Free Software Foundation; either version 2.1 of
   the License, or (at your option) any later version.

   The m17n library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with the m17n library; if not, write to the Free
   Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
   02111-1307, USA.  */

#include <stdlib.h>
#include <string.h>
#include <libintl.h>
#include <m17n.h>
#include <m17n-misc.h>
#include <gtk/gtk.h>
#include <config.h>
#include "m17n-im-config.h"

#define _(String) dgettext (PACKAGE, String)

#define CONFIG_CALLBACK_DATA " config-callback-data"
#define CONFIG_STATUS_DATA " config-status-data"
#define CONFIG_TREE_VIEW "  config-tree-view"

typedef void (*MimConfigCallbackFunc) (GtkWidget *widget, gpointer data);

typedef struct _MimConfigCallback
{
  GtkWidget *widget;
  MimConfigCallbackFunc func;
  gpointer data;
} MimConfigCallback;

typedef struct _MimConfigStatus
{
  /* Number of available input methods.  */
  gint num_im;
  /* Number of modified input methods.  */
  gint num_modified;
} MimConfigStatus;

/* Status of variables and commands of an input method.  */

enum MimStatus
  {
    MIM_STATUS_DEFAULT,
    MIM_STATUS_CUSTOMIZED,
    MIM_STATUS_MODIFIED,
    MIM_STATUS_NO,
    MIM_STATUS_MAX
  };

static char *mim_status_str[MIM_STATUS_MAX];

enum MimStatus
get_mim_status (MSymbol lang, MSymbol name)
{
  MPlist *plist;
  enum MimStatus status = MIM_STATUS_NO;

  for (plist = minput_get_variable (lang, name, Mnil);
       plist && mplist_key (plist) != Mnil; plist = mplist_next (plist))
    {
      MPlist *p = mplist_value (plist);
      MSymbol status_symbol;

      status = MIM_STATUS_DEFAULT;
      p = mplist_next (mplist_next (p));
      status_symbol = mplist_value (p);
      if (status_symbol != Mnil && status_symbol != Minherited)
	return (status_symbol == Mcustomized
		? MIM_STATUS_CUSTOMIZED : MIM_STATUS_MODIFIED);
    }
  for (plist = minput_get_command (lang, name, Mnil);
       plist && mplist_key (plist) != Mnil; plist = mplist_next (plist))
    {
      MPlist *p = mplist_value (plist);
      MSymbol status_symbol;

      status = MIM_STATUS_DEFAULT;
      p = mplist_next (mplist_next (p));
      status_symbol = mplist_value (p);
      if (status_symbol != Mnil && status_symbol != Minherited)
	return (status_symbol == Mcustomized
		? MIM_STATUS_CUSTOMIZED : MIM_STATUS_MODIFIED);
    }
  return status;
}

/* Columns of each row.  */
enum
  {
    /* parent: language name
        child: IM name  */
    COL_TAG = 0,
    /* parent: NULL or "modified"
        child: "default", "customized", or "modified"  */
    COL_STATUS_STR,
    /* parent: num of modified children
        child: enum MimStatus */
    COL_STATUS,
    /* parent: Mnil
        child: symbolic language name.  */
    COL_LANG,
    /* parent: Mnil
        child: symbolic IM name.  */
    COL_NAME,
    /* number of columns  */
    NUM_COLS
  };

/* Called when a row is expanded.  We may have to initialize
   children.  */
static void
tree_expanded_cb (GtkTreeView *tree, GtkTreeIter *parent,
		  GtkTreePath *path, gpointer data)
{
  GtkTreeModel *model;
  GtkTreeIter iter;
  MSymbol lang, name;

  model = gtk_tree_view_get_model (tree);
  if (gtk_tree_model_iter_children (model, &iter, parent))
    {
      gchar *status_str;

      gtk_tree_model_get (model, &iter, COL_STATUS_STR, &status_str, -1);
      if (! status_str)
	{
	  /* The first child is not yet initialized, and that means
	     the remaining children are not initialized either.  */
	  gtk_tree_model_get (model, &iter, COL_LANG, &lang, -1);
	  do {
	    enum MimStatus im_status;

	    gtk_tree_model_get (model, &iter, COL_NAME, &name, -1);
	    im_status = get_mim_status (lang, name);
	    gtk_tree_store_set (GTK_TREE_STORE (model), &iter,
				COL_STATUS_STR, mim_status_str[im_status],
				COL_STATUS, im_status,
				-1);
	  } while (gtk_tree_model_iter_next (model, &iter));
	}
    }
}

static void
edit_im (GtkTreeView *tree, MSymbol lang, MSymbol name)
{
  GtkWidget *dialog, *frame, *table;
  GtkTooltips *tip = gtk_tooltips_new ();
  gint response;
  MPlist *plist;

  dialog = (gtk_dialog_new_with_buttons
	    (msymbol_name (name),
	     GTK_WINDOW (gtk_widget_get_toplevel (GTK_WIDGET (tree))),
	     GTK_DIALOG_DESTROY_WITH_PARENT | GTK_DIALOG_NO_SEPARATOR,
	     _("Default"), 0,
	     GTK_STOCK_REVERT_TO_SAVED, GTK_RESPONSE_NO,
	     GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
	     NULL));

  frame = gtk_frame_new (_("Variables"));
  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->vbox), frame,
		      FALSE, FALSE, 0);
  create_variable_entries (frame, tip, lang, name);

  frame = gtk_frame_new (_("Commands"));
  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->vbox), frame,
		      FALSE, FALSE, 0);
  create_command_entries (frame, tip, lang, name);

  gtk_widget_show_all (dialog);
  response = gtk_dialog_run (GTK_DIALOG (dialog));
  if (response == 0)
    printf ("Default\n");
  else if (response == GTK_RESPONSE_NO)
    printf ("GTK_RESPONSE_NO\n");
  else if (response == GTK_RESPONSE_CANCEL)
    printf ("GTK_RESPONSE_CANCEL\n");
  gtk_widget_destroy (dialog);
}

static void
update_child_row (GtkTreeModel *model, GtkTreeIter *iter,
		  enum MimStatus status, MimConfigStatus *config_status,
		  GtkTreeView *tree)
{
  GtkTreeIter parent;
  gint inc_modified;

  inc_modified = (status == MIM_STATUS_MODIFIED ? 1 : -1);

  gtk_tree_store_set (GTK_TREE_STORE (model), iter,
		      COL_STATUS_STR, mim_status_str[status],
		      COL_STATUS, status, -1);
  if (gtk_tree_model_iter_parent (model, &parent, iter))
    {
      gint num_modified;
      gchar *status_str;

      gtk_tree_model_get (model, &parent, COL_STATUS, &num_modified, -1);
      num_modified += inc_modified;
      gtk_tree_store_set (GTK_TREE_STORE (model), &parent,
			  COL_STATUS, num_modified, -1);
      if (num_modified <= 1)
	{
	  status_str = (status == MIM_STATUS_MODIFIED
			? mim_status_str[MIM_STATUS_MODIFIED] : NULL);
	  gtk_tree_store_set (GTK_TREE_STORE (model), &parent,
			      COL_STATUS_STR, status_str);
	}
    }
      
  if (! config_status)
    config_status = g_object_get_data (G_OBJECT (model), CONFIG_STATUS_DATA);
  config_status->num_modified += inc_modified;
  if (tree && config_status->num_modified <= 1)
    {
      MimConfigCallback *callback;

      callback = g_object_get_data (G_OBJECT (tree), CONFIG_CALLBACK_DATA);
      if (callback)
	callback->func (callback->widget, callback->data);
    }
}

static void
tree_activated_cb (GtkTreeView *tree, GtkTreePath *path,
		   GtkTreeViewColumn *column, gpointer data)
{
  GtkTreeModel *model;
  GtkTreeIter iter;

  model = gtk_tree_view_get_model (tree);
  if (gtk_tree_model_get_iter (model, &iter, path))
    {
      MSymbol lang, name;

      gtk_tree_model_get (model, &iter, COL_LANG, &lang, COL_NAME, &name, -1);
      if (lang != Mnil)
	{
	  /* child row for an IM */
	  enum MimStatus old, new;

	  old = get_mim_status (lang, name);
	  edit_im (tree, lang, name);
	  new = get_mim_status (lang, name);
	  if (old != new)
	    update_child_row (model, &iter, new, NULL, tree);
	}
      else
	{
	  /* parent row for a language */
	  if (gtk_tree_view_row_expanded (tree, path))
	    gtk_tree_view_collapse_row (tree, path);
	  else
	    gtk_tree_view_expand_row (tree, path, TRUE);
	}
    }
}

typedef struct _MimTable
{
  gchar *lang;
  gchar *name;
  MSymbol symlang;
  MSymbol symname;
} MimTable;

static int
sort_im (const void *p1, const void *p2)
{
  const MimTable *t1 = p1;
  const MimTable *t2 = p2;
  int result = strcmp (t1->lang, t2->lang);

  return (result ? result : strcmp (t1->name, t2->name));
}

static GtkTreeStore *
make_store_for_input_methods ()
{
  GtkTreeStore *store;
  MPlist *imlist, *p;
  int i;
  MimTable *imtable;
  char *lang;
  GtkTreeIter iter1, iter2;
  enum MimStatus status;
  MimConfigStatus *config_status;

  store = gtk_tree_store_new (NUM_COLS,
			      G_TYPE_STRING,  /* COL_TAG */
			      G_TYPE_STRING,  /* COL_STATUS_STR */
			      G_TYPE_UINT,    /* COL_STATUS */
			      G_TYPE_POINTER, /* COL_LANG */
			      G_TYPE_POINTER  /* COL_NAME */
			      );

  config_status = g_new0 (MimConfigStatus, 1);
  gtk_tree_store_append (store, &iter1, NULL);
  status = get_mim_status (Mt, Mnil);
  gtk_tree_store_set (store, &iter1,
		      COL_TAG, _("global"),
		      COL_STATUS_STR, mim_status_str[status],
		      COL_STATUS, status, 
		      COL_LANG, Mt,
		      COL_NAME, Mnil,
		      -1);

  imlist = mdatabase_list (msymbol ("input-method"), Mnil, Mnil, Mnil);
  config_status->num_im = mplist_length (imlist);
  imtable = g_newa (MimTable, config_status->num_im);
  for (i = 0, p = imlist; mplist_key (p) != Mnil; p = mplist_next (p))
    {
      MDatabase *mdb = (MDatabase *) mplist_value (p);
      MSymbol *tag = mdatabase_tag (mdb);

      if (tag[1] != Mnil && tag[2] != Mnil)
	{
	  MSymbol language = mlanguage_name (tag[1]);

	  if (language != Mnil)
	    imtable[i].lang = msymbol_name (language);
	  else
	    /* `~' is for putting this element at the tail by sort.  */
	    imtable[i].lang = "~other";
	  imtable[i].name = msymbol_name (tag[2]);
	  imtable[i].symlang = tag[1];
	  imtable[i].symname = tag[2];
	  i++;
	}
    }
  m17n_object_unref (imlist);
  config_status->num_im = i;
  qsort (imtable, config_status->num_im, sizeof (MimTable), sort_im);

  for (lang = NULL, i = 0; i < config_status->num_im; i++)
    {
      if (lang != imtable[i].lang)
	{
	  gchar *name;

	  gtk_tree_store_append (store, &iter1, NULL);
	  lang = imtable[i].lang;
	  if (lang[0] != '~')
	    {
	      MText *native_text;
	      gchar *native = NULL;
	      int nbytes;

	      if (imtable[i].symlang != Mt
		  && (native_text = mlanguage_text (imtable[i].symlang)))
		{
		  enum MTextFormat fmt;

		  native = mtext_data (native_text, &fmt, &nbytes,
				       NULL, NULL);
		  if (fmt != MTEXT_FORMAT_US_ASCII
		      && fmt != MTEXT_FORMAT_UTF_8)
		    native = 0;
		}
	      if (native)
		{
		  name = alloca (strlen (lang) + nbytes + 4);
		  sprintf (name, "%s (%s)", lang, native);
		}
	      else
		name = lang;
	    }
	  else
	    name = lang + 1;

	  gtk_tree_store_set (store, &iter1,
			      COL_TAG, name,
			      COL_STATUS_STR, NULL,
			      COL_STATUS, 0,
			      COL_LANG, Mnil,
			      COL_NAME, Mnil,
			      -1);
	}
      gtk_tree_store_append (store, &iter2, &iter1);
      gtk_tree_store_set (store, &iter2,
			  COL_TAG, imtable[i].name,
			  COL_STATUS_STR, NULL,
			  COL_LANG, imtable[i].symlang,
			  COL_NAME, imtable[i].symname,
			  -1);
    }
  config_status->num_modified = 0;
  g_object_set_data_full (G_OBJECT (store), CONFIG_STATUS_DATA,
			  config_status, g_free);
  return store;
}

static gboolean
revert_to_saved (GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter,
	       gpointer data)
{
  enum MimStatus status;
  MSymbol lang, name;
  MimConfigStatus *config_status = data;

  gtk_tree_model_get (model, iter, COL_LANG, &lang, COL_NAME, &name, -1);
  if (lang == Mnil)
    return FALSE;
  gtk_tree_model_get (model, iter, COL_STATUS, &status, -1);  
  if (status != MIM_STATUS_MODIFIED)
    return FALSE;
  minput_config_variable (lang, name, Mnil, NULL);
  minput_config_command (lang, name, Mnil, NULL);
  status = get_mim_status (lang, name);
  update_child_row (model, iter, status, config_status, NULL);
  return FALSE;
}

static gboolean
set_as_saved (GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter,
	   gpointer data)
{
  enum MimStatus status;
  MSymbol lang, name;
  MimConfigStatus *config_status = data;

  gtk_tree_model_get (model, iter, COL_LANG, &lang, COL_NAME, &name, -1);
  if (lang == Mnil)
    return FALSE;
  gtk_tree_model_get (model, iter, COL_STATUS, &status, -1);  
  if (status != MIM_STATUS_MODIFIED)
    return FALSE;
  status = get_mim_status (lang, name);
  update_child_row (model, iter, status, config_status, NULL);
  return FALSE;
}

static int initialized = 0;

static void
destroy_cb (GtkWidget *widget, gpointer data)
{
  M17N_FINI ();
}


/* Public API */

GtkWidget *
mim_config_new (GCallback func, gpointer data)
{
  GtkWidget *tree, *config;
  GtkTreeStore *store;
  GtkCellRenderer *renderer;
  GtkTreeViewColumn *column;

  if (initialized)
    return NULL;
  M17N_INIT ();
  if (merror_code < 0)
    return NULL;

#if ENABLE_NLS
  bindtextdomain ("m17n-im-config", GETTEXTDIR);
  bind_textdomain_codeset ("m17n-im-config", "UTF-8");
#endif

  mim_status_str[MIM_STATUS_DEFAULT] = _("default");
  mim_status_str[MIM_STATUS_CUSTOMIZED] = _("customized");
  mim_status_str[MIM_STATUS_MODIFIED] = _("modified");
  mim_status_str[MIM_STATUS_NO] = _("uncustomizable");

  store = make_store_for_input_methods ();
  tree = gtk_tree_view_new_with_model (GTK_TREE_MODEL (store));
  g_object_unref (G_OBJECT (store));

  renderer = gtk_cell_renderer_text_new ();
  column = (gtk_tree_view_column_new_with_attributes
	    (_("Input Method"), renderer, "text", COL_TAG, NULL));
  gtk_tree_view_append_column (GTK_TREE_VIEW (tree), column);

  renderer = gtk_cell_renderer_text_new ();
  column = (gtk_tree_view_column_new_with_attributes
	    (_("status"), renderer, "text", COL_STATUS_STR, NULL));
  gtk_tree_view_append_column (GTK_TREE_VIEW (tree), column);

  g_signal_connect (G_OBJECT (tree), "row-expanded",
		    G_CALLBACK (tree_expanded_cb), NULL);
  g_signal_connect (G_OBJECT (tree), "row-activated",
		    G_CALLBACK (tree_activated_cb), NULL);

  config =gtk_scrolled_window_new (NULL, NULL);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (config),
				  GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
  gtk_scrolled_window_add_with_viewport (GTK_SCROLLED_WINDOW (config), tree);
  g_signal_connect (G_OBJECT (config), "destroy",
		    G_CALLBACK (destroy_cb), NULL);

  g_object_set_data (G_OBJECT (config), CONFIG_TREE_VIEW, tree);
  if (func)
    {
      MimConfigCallback *callback;

      callback = g_new (MimConfigCallback, 1);
      callback->widget = config;
      callback->func = (MimConfigCallbackFunc) func;
      callback->data = data;
      g_object_set_data_full (G_OBJECT (tree), CONFIG_CALLBACK_DATA,
			      callback, g_free);
    }

  return config;
}

gboolean
mim_config_modified (GtkWidget *config)
{
  GtkTreeView *tree;
  GtkTreeModel *model;
  MimConfigStatus *config_status;

  tree = g_object_get_data (G_OBJECT (config), CONFIG_TREE_VIEW);
  model = gtk_tree_view_get_model (tree);
  config_status = g_object_get_data (G_OBJECT (model), CONFIG_STATUS_DATA);

  return (config_status->num_modified > 0 ? TRUE : FALSE);
}

gboolean
mim_config_revert (GtkWidget *config)
{
  GtkTreeView *tree;
  GtkTreeModel *model;
  MimConfigStatus *config_status;

  tree = g_object_get_data (G_OBJECT (config), CONFIG_TREE_VIEW);
  model = gtk_tree_view_get_model (tree);
  config_status = g_object_get_data (G_OBJECT (model), CONFIG_STATUS_DATA);

  if (config_status->num_modified == 0)
    return FALSE;
  gtk_tree_model_foreach (model, revert_to_saved, config_status);
  return TRUE;
}

gboolean
mim_config_save (GtkWidget *config)
{
  GtkTreeView *tree;
  GtkTreeModel *model;
  MimConfigStatus *config_status;

  tree = g_object_get_data (G_OBJECT (config), CONFIG_TREE_VIEW);
  model = gtk_tree_view_get_model (tree);
  config_status = g_object_get_data (G_OBJECT (model), CONFIG_STATUS_DATA);

  if (config_status->num_modified == 0)
    return FALSE;
  minput_save_config ();
  gtk_tree_model_foreach (model, set_as_saved, config_status);
  return TRUE;
}

GtkTreeView *
mim_config_get_tree_view (GtkWidget *config)
{
  GtkTreeView *tree;

  tree = g_object_get_data (G_OBJECT (config), CONFIG_TREE_VIEW);
  return tree;
}
